#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import requests
import urllib

class Endpoint:

	"""
	GlassFish domain management.
	
	:param str host:
	   Endpoint host.
	:param int port:
	   Endpoint port.
	:param str username:
	   Endpoint username.
	:param str password:
	   Endpoint password.
	"""
	
	def __init__(self, host, port, username, password):
	
		self.__base_uri = "https://{}:{}/management/domain".format(host, port)
		self.__username = username
		self.__password = password
		
	def __res_uri(self, path):
	
		return "{}{}".format(self.__base_uri, path)
		
	def __headers(self):
	
		return {
			"Accept": "application/json",
			"X-Requested-By": "GlassFish REST HTML interface"
		}
		
	def __auth(self):
	
		return requests.auth.HTTPBasicAuth(self.__username, self.__password)
		
	def get(self, res_path, timeout=10.):
	
		"""
		Get operation.
		
		:param str res_path:
		   Resource path.
		:param float timeout:
		   Timeout in seconds.
		:rtype:
		   tuple
		:return:
		   Tuple with status code and response body.
		"""
		
		resp = requests.get(
			self.__res_uri(res_path),
			headers=self.__headers(),
			verify=False,
			auth=self.__auth(),
			timeout=timeout
		)
		return (
			resp.status_code,
			json.loads(resp.text) if resp.status_code == 200 else {}
		)
		
	def post(self, res_path, data=None, files=None, timeout=10.):
	
		"""
		Post operation.
		
		:param str res_path:
		   Resource path.
		:param list data:
		   Request parameters for data.
		:param list files:
		   Request parameters for files.
		:param float timeout:
		   Timeout in seconds.
		:rtype:
		   tuple
		:return:
		   Tuple with status code and response body.
		"""
		
		resp = requests.post(
			self.__res_uri(res_path),
			data=data,
			files=files,
			headers=self.__headers(),
			verify=False,
			auth=self.__auth(),
			timeout=timeout
		)
		return (
			resp.status_code,
			json.loads(resp.text)
		)
		
	def delete(self, res_path, timeout=10.):
	
		"""
		Delete operation.
		
		:param str res_path:
		   Resource path.
		:param float timeout:
		   Timeout in seconds.
		:rtype:
		   tuple
		:return:
		   Tuple with status code and response body.
		"""
		
		resp = requests.delete(
			self.__res_uri(res_path),
			headers=self.__headers(),
			verify=False,
			auth=self.__auth(),
			timeout=timeout
		)
		return (
			resp.status_code,
			json.loads(resp.text) if resp.status_code == 200 else {}
		)
		
class Domain:

	"""
	Domain for an endpoint. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param float avail_timeout:
	   Availability check timeout in seconds.
	"""
	
	def __init__(self, endpoint, avail_timeout):
	
		self.__endpoint = endpoint
		self.__avail_timeout = avail_timeout
		self.__available = None
		
	def available(self, context):
	
		"""
		Domain availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				status_code, msg = self.__endpoint.get("")
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
		
	def insert(self, context):
	
		"""
		Wait domain to be available.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		:raise Exception:
		   Domain could not been inserted.
		"""
		
		status_code, msg = self.__endpoint.get("", self.__avail_timeout)
		if status_code != 200:
			raise Exception("GlassFish domain could not been inserted")
		self.__available = True
		
	def delete(self, context):
	
		"""
		Does nothing.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.__available = None
		
	def application(self, name, context_root, path):
	
		"""
		Domain application.
		
		:param str name:
		   Application name.
		:param str context_root:
		   Appliaction context root. May be ``None``.
		:param Contextual path:
		   File path.
		:rtype:
		   Application
		"""
		
		return Application(self.__endpoint, name, context_root, path)
		
	def jdbc_resource(self, name, pool_name):
	
		"""
		Domain JDBC resource.
		
		:param str name:
		   Resource name.
		:param str pool_name:
		   Resource pool name.
		:rtype:
		   JDBCResource
		"""
		
		return JDBCResource(self.__endpoint, name, pool_name)
		
	def connector_resource(self, name, pool_name):
	
		"""
		Domain connector resource.
		
		:param str name:
		   Resource name.
		:param str pool_name:
		   Resource pool name.
		:rtype:
		   ConnectorResource
		"""
		
		return ConnectorResource(self.__endpoint, name, pool_name)
		
	def mail_session(self, name, host, username, mail_from, props):
	
		"""
		Domain mail session.
		
		:param str name:
		   Mail session name.
		:param str host:
		   Mail host.
		:param str username:
		   Mail username.
		:param str mail_from:
		   Mail "from" address.
		:param dict props:
		   Extra properties.
		:rtype:
		   MailSession
		"""
		
		return MailSession(self.__endpoint, name, host, username, mail_from,
				props)
		
	def custom_resource(self, name, restype, factclass, props):
	
		"""
		Domain custom resource.
		
		:param str name:
		   Resource name.
		:param str restype:
		   Resource type.
		:param str factclass:
		   Resource factory class.
		:param dict props:
		   Resource properties.
		:rtype:
		   CustomResource
		"""
		
		return CustomResource(self.__endpoint, name, restype, factclass, props)
		
	def jdbc_connection_pool(self, name, res_type, ds_classname, props):
			
		"""
		Domain JDBC connection pool.
		
		:param str name:
		   Resource name.
		:param str res_type:
		   Resource type.
		:param str ds_classname:
		   Data source class name.
		:param dict props:
		   Connection pool properties.
		:rtype:
		   JDBCConnectionPool
		"""
		
		return JDBCConnectionPool(self.__endpoint, name, res_type,
				ds_classname, props)
				
	def connector_connection_pool(self, name, res_adapter_name, conn_def_name,
			props):
			
		"""
		Domain connector connection pool.
		
		:param str name:
		   Resource name.
		:param str res_adapter_name:
		   Resource adapter name.
		:param str conn_def_name:
		   Resource connection definition name.
		:param dict props:
		   Connection pool properties.
		:rtype:
		   ConnectorConnectionPool
		"""
		
		return ConnectorConnectionPool(self.__endpoint, name, res_adapter_name,
				conn_def_name, props)
		
class Application:

	"""
	Domain application. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param str name:
	   Application name.
	:param str context_root:
	   Appliaction context root. May be ``None``.
	:param Contextual path:
	   File path.
	"""
	
	def __init__(self, endpoint, name, context_root, path):
	
		self.__endpoint = endpoint
		self.__name = name
		self.__context_root = context_root
		self.__path = path
		self.__available = None
		
	def available(self, context):
	
		"""
		Application availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				encoded_name = urllib.parse.quote_plus(self.__name)
				status_code, msg = self.__endpoint.get(
					"/applications/application/{}".format(encoded_name)
				)
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
		
	def insert(self, context):
	
		"""
		Deploy application.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		module_file = open(context.resolve(self.__path), "rb")
		data = {
			"name": self.__name
		}
		if self.__context_root is not None:
			data["contextroot"] = self.__context_root
		status_code, msg = self.__endpoint.post(
			"/applications/application",
			data=data,
			files={
				"id": module_file
			},
			timeout=60.
		)
		module_file.close()
		self.__available = True
		
	def delete(self, context):
	
		"""
		Undeploy application.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		status_code, msg = self.__endpoint.delete(
			"/applications/application/{}".format(self.__name)
		)
		self.__available = False
		
class JDBCResource:

	"""
	JDBC resource. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param str name:
	   Resource name.
	:param str pool_name:
	   Resource pool name.
	"""
	
	def __init__(self, endpoint, name, pool_name):
	
		self.__endpoint = endpoint
		self.__name = name
		self.__pool_name = pool_name
		self.__available = None
		
	def available(self, context):
	
		"""
		Resource availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				encoded_name = urllib.parse.quote_plus(self.__name)
				status_code, msg = self.__endpoint.get(
					"/resources/jdbc-resource/{}".format(encoded_name)
				)
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
		
	def insert(self, context):
	
		"""
		Create resource.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		status_code, msg = self.__endpoint.post(
			"/resources/jdbc-resource",
			data={
				"id": self.__name,
				"poolName": self.__pool_name
			}
		)
		self.__available = True
		
	def delete(self, context):
	
		"""
		Remove resource.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		encoded_name = urllib.parse.quote_plus(self.__name)
		status_code, msg = self.__endpoint.delete(
			"/resources/jdbc-resource/{}".format(encoded_name)
		)
		self.__available = False
		
class ConnectorResource:

	"""
	Connector resource. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param str name:
	   Resource name.
	:param str pool_name:
	   Resource pool name.
	"""
	
	def __init__(self, endpoint, name, pool_name):
	
		self.__endpoint = endpoint
		self.__name = name
		self.__pool_name = pool_name
		self.__available = None
		
	def available(self, context):
	
		"""
		Resource availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				encoded_name = urllib.parse.quote_plus(self.__name)
				status_code, msg = self.__endpoint.get(
					"/resources/connector-resource/{}".format(encoded_name)
				)
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
		
	def insert(self, context):
	
		"""
		Create resource.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		status_code, msg = self.__endpoint.post(
			"/resources/connector-resource",
			data={
				"id": self.__name,
				"poolname": self.__pool_name
			}
		)
		self.__available = True
		
	def delete(self, context):
	
		"""
		Remove resource.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		encoded_name = urllib.parse.quote_plus(self.__name)
		status_code, msg = self.__endpoint.delete(
			"/resources/connector-resource/{}".format(encoded_name)
		)
		self.__available = False
		
class MailSession:

	"""
	Mail session. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param str name:
	   Mail session name.
	:param str host:
	   Mail host.
	:param str username:
	   Mail username.
	:param str mail_from:
	   Mail "from" address.
	:param dict props:
	   Extra properties.
	"""
	
	def __init__(self, endpoint, name, host, username, mail_from, props):
	
		self.__endpoint = endpoint
		self.__name = name
		self.__host = host
		self.__username = username
		self.__mail_from = mail_from
		self.__props = props
		self.__available = None
		
	def available(self, context):
	
		"""
		Mail session availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				encoded_name = urllib.parse.quote_plus(self.__name)
				status_code, msg = self.__endpoint.get(
					"/resources/mail-resource/{}".format(encoded_name)
				)
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
		
	def insert(self, context):
	
		"""
		Create mail session.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		status_code, msg = self.__endpoint.post(
			"/resources/mail-resource",
			data={
				"id": self.__name,
				"host": self.__host,
				"user": self.__username,
				"from": self.__mail_from,
				"property": props_value(self.__props)
			}
		)
		self.__available = True
		
	def delete(self, context):
	
		"""
		Remove mail session.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		encoded_name = urllib.parse.quote_plus(self.__name)
		status_code, msg = self.__endpoint.delete(
			"/resources/mail-resource/{}".format(encoded_name)
		)
		self.__available = False
		
class CustomResource:

	"""
	Domain resource. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param str name:
	   Resource name.
	:param str restype:
	   Resource type.
	:param str factclass:
	   Resource factory class.
	:param dict props:
	   Resource properties.
	"""
	
	def __init__(self, endpoint, name, restype, factclass, props):
	
		self.__endpoint = endpoint
		self.__name = name
		self.__restype = restype
		self.__factclass = factclass
		self.__props = props
		self.__available = None
		
	def available(self, context):
	
		"""
		Resource availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				encoded_name = urllib.parse.quote_plus(self.__name)
				status_code, msg = self.__endpoint.get(
					"/resources/custom-resource/{}".format(encoded_name)
				)
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
		
	def insert(self, context):
	
		"""
		Create resource.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		status_code, msg = self.__endpoint.post(
			"/resources/custom-resource",
			data={
				"id": self.__name,
				"restype": self.__restype,
				"factoryclass": self.__factclass,
				"property": props_value(self.__props)
			}
		)
		self.__available = True
		
	def delete(self, context):
	
		"""
		Remove resource.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		encoded_name = urllib.parse.quote_plus(self.__name)
		status_code, msg = self.__endpoint.delete(
			"/resources/custom-resource/{}".format(encoded_name)
		)
		self.__available = False
		
class JDBCConnectionPool:

	"""
	JDBC connection pool. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param str name:
	   Resource name.
	:param str res_type:
	   Resource type.
	:param str ds_classname:
	   Data source class name.
	:param dict props:
	   Connection pool properties.
	"""
	
	def __init__(self, endpoint, name, res_type, ds_classname, props):
	
		self.__endpoint = endpoint
		self.__name = name
		self.__res_type = res_type
		self.__ds_classname = ds_classname
		self.__props = props
		self.__available = None
		
	def available(self, context):
	
		"""
		Connection pool availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				encoded_name = urllib.parse.quote_plus(self.__name)
				status_code, msg = self.__endpoint.get(
					"/resources/jdbc-connection-pool/{}".format(encoded_name)
				)
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
		
	def insert(self, context):
	
		"""
		Create connection pool.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		status_code, msg = self.__endpoint.post(
			"/resources/jdbc-connection-pool",
			data={
				"id": self.__name,
				"resType": self.__res_type,
				"datasourceClassname": self.__ds_classname,
				"property": props_value(self.__props)
			}
		)
		self.__available = True
		
	def delete(self, context):
	
		"""
		Remove connection pool.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		encoded_name = urllib.parse.quote_plus(self.__name)
		status_code, msg = self.__endpoint.delete(
			"/resources/jdbc-connection-pool/{}".format(encoded_name)
		)
		self.__available = False
		
class ConnectorConnectionPool:

	"""
	Connector connection pool. Implements :class:`Component`.
	
	:param Endpoint endpoint:
	   Domain endpoint.
	:param str name:
	   Resource name.
	:param str res_adapter_name:
	   Resource adapter name.
	:param str conn_def_name:
	   Resource connection definition name.
	:param dict props:
	   Connection pool properties.
	"""
	
	def __init__(self, endpoint, name, res_adapter_name, conn_def_name, props):
	
		self.__endpoint = endpoint
		self.__name = name
		self.__res_adapter_name = res_adapter_name
		self.__conn_def_name = conn_def_name
		self.__props = props
		self.__available = None
		
	def available(self, context):
	
		"""
		Connection pool availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			if self.__available is None:
				encoded_name = urllib.parse.quote_plus(self.__name)
				status_code, msg = self.__endpoint.get(
					"/resources/connector-connection-pool/{}".format(
							encoded_name)
				)
				self.__available = status_code == 200
		except:
			self.__available = False
			
		return self.__available
			
	def insert(self, context):
	
		"""
		Create connection pool.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		status_code, msg = self.__endpoint.post(
			"/resources/connector-connection-pool",
			data={
				"id": self.__name,
				"resourceAdapterName": self.__res_adapter_name,
				"connectiondefinitionname": self.__conn_def_name,
				"property": props_value(self.__props)
			}
		)
		self.__available = True
		
	def delete(self, context):
	
		"""
		Remove connection pool.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		encoded_name = urllib.parse.quote_plus(self.__name)
		status_code, msg = self.__endpoint.delete(
			"/resources/connector-connection-pool/{}".format(encoded_name)
		)
		self.__available = False
		
def domain(host, port, username, password, avail_timeout=240.):
	
	"""
	Endpoint domain.
	
	:param str host:
	   Endpoint host.
	:param int port:
	   Endpoint port.
	:param str username:
	   Endpoint username.
	:param str password:
	   Endpoint password.
	:param float avail_timeout:
	   Availability check timeout in seconds.
	:rtype:
	   Domain
	:return:
	   Domain for the given endpoint parameters.
	"""
	
	return Domain(Endpoint(host, port, username, password), avail_timeout)
	
def props_value(props):

	"""
	Properties value.
	
	:param dict props:
	   Properties dictionary.
	:rtype:
	   str
	:return:
	   Properties as string.
	"""
	
	sep = ""
	result_value = ""
	for prop_key, prop_value in props.items():
		result_value = "{}{}{}={}".format(
			result_value,
			sep,
			prop_key,
			prop_value
		)
		sep = ":"
	return result_value




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

import io
import os
import subprocess

class VagrantObject:

	"""
	Vagrant object.
	"""
	
	def __init__(self):
	
		pass
		
	def read(self, cmd_args):
	
		"""
		Execute Vagrant read command.
		
		:param list cmd_args:
		   Command argument list.
		"""
		
		args = [
			"vagrant",
			"--machine-readable"
		]
		args.extend(cmd_args)
		proc = subprocess.Popen(args, stdout=subprocess.PIPE)
		for line in proc.stdout.readlines():
			if len(line) == 0:
				break
			yield line.decode("UTF-8").rsplit(",")
		proc.wait()
		
	def write(self, cmd_args):
	
		"""
		Execute Vagrant write command.
		
		:param list cmd_args:
		   Command argument list.
		"""
		
		args = [
			"vagrant"
		]
		args.extend(cmd_args)
		subprocess.call(args)
		
class BoxFile(VagrantObject):

	"""
	Vagrant box from file. Implements :class:`Component`.
	
	:param Contextual name:
	   Box name.
	:param Contextual image_dir:
	   Box image directory.
	"""
	
	def __init__(self, name, image_dir):
	
		self.__name = name
		self.__image_dir = image_dir
		self.__available = None
		
	def __box_list(self):
	
		result_entry = {}
		for line in self.read([
			"box",
			"list"
		]):
			key = line[2]
			value = line[3].strip()
			if key == "box-name":
				if len(result_entry) > 0:
					yield result_entry
					result_entry = {}
				result_entry["name"] = value
			elif key == "box-provider":
				result_entry["provider"] = value
			elif key == "box-version":
				result_entry["version"] = value
		if len(result_entry) > 0:
			yield result_entry
			
	def __path(self, context):
	
		image_dir = context.resolve(self.__image_dir)
		for fname in os.listdir(image_dir):
			if fname.endswith(".box"):
				return os.path.join(image_dir, fname)
		raise Exception("Box image not found at {}".format(image_dir))
		
	def available(self, context):
	
		"""
		Box is available for the calling user.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		if self.__available is None:
			avail = False
			name = context.resolve(self.__name)
			for box in self.__box_list():
				if box["name"] == name and box["version"] == "0":
					avail = True
					break
			self.__available = avail
		return self.__available
		
	def insert(self, context):
	
		"""
		Add Vagrant box to the calling user.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.write([
			"box",
			"add",
			"--name", context.resolve(self.__name),
			self.__path(context)
		])
		
	def delete(self, context):
	
		"""
		Delete Vagrant box from the calling user.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.write([
			"box",
			"remove",
			context.resolve(self.__name)
		])
		
class Instance(VagrantObject):

	"""
	Vagrant instance.
	
	:param Contextual config_dir:
	   Instance configuration directory.
	:param Contextual name:
	   Instance name.
	"""
	
	def __init__(self, config_dir, name):
	
		self.__config_dir = config_dir
		self.__name = name
		
	def read(self, context, cmd_args):
	
		"""
		Execute Vagrant read command on instance placed into configuration
		directory.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		:param list cmd_args:
		   Command argument list.
		"""
		
		try:
			current_dir = os.getcwd()
			os.chdir(context.resolve(self.__config_dir))
			args = []
			args.extend(cmd_args)
			args.append(context.resolve(self.__name))
			yield from super().read(args)
		finally:
			os.chdir(current_dir)
			
	def write(self, context, cmd_args):
	
		"""
		Execute Vagrant write command on instance placed into configuration
		directory.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		:param list cmd_args:
		   Command argument list.
		"""
		
		try:
			current_dir = os.getcwd()
			os.chdir(context.resolve(self.__config_dir))
			args = []
			args.extend(cmd_args)
			args.append(context.resolve(self.__name))
			super().write(args)
		finally:
			os.chdir(current_dir)
			
	def state(self, context):
	
		"""
		Get instance state.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		:rtype:
		   str
		:return:
		   Instance state name.
		"""
		
		state = None
		for line in self.read(context, [
			"status",
			context.resolve(self.__name)
		]):
			if line[2] == "state":
				state = line[3].strip()
		return state
			
	def created(self):
	
		"""
		Instance created component.
		
		:rtype:
		   InstanceCreated
		"""
		
		return InstanceCreated(self)
		
	def running(self):
	
		"""
		Instance running component.
		
		:rtype:
		   InstanceRunning
		"""
		
		return InstanceRunning(self)
		
class InstanceCreated:

	"""
	Vagrant instance created. Implements :class:`Component`.
	
	:param Instance inst:
	   Managed instance.
	"""
	
	def __init__(self, inst):
	
		self.__inst = inst
		self.__available = None
		
	def available(self, context):
	
		"""
		Check availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		:rtype:
		   bool
		:return:
		   Availability value.
		"""
		if self.__available is None:
			state = self.__inst.state(context)
			self.__available = state is not None and state != "not_created"
		return self.__available
		
	def insert(self, context):
	
		"""
		Create instance.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.__inst.write(context, [
			"up"
		])
		
	def delete(self, context):
	
		"""
		Delete instance.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.__inst.write(context, [
			"destroy",
			"-f"
		])
		
class InstanceRunning:

	"""
	Vagrant instance running. Implements :class:`Component`.
	
	:param Instance inst:
	   Managed instance.
	"""
	
	def __init__(self, inst):
	
		self.__inst = inst
		self.__available = None
		
	def available(self, context):
	
		"""
		Check availability.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		:rtype:
		   bool
		:return:
		   Availability value.
		"""
		
		if self.__available is None:
			self.__available = self.__inst.state(context) == "running"
		return self.__available
		
	def insert(self, context):
	
		"""
		Run instance.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.__inst.write(context, [
			"up"
		])
		
	def delete(self, context):
	
		"""
		Stop instance.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.__inst.write(context, [
			"halt"
		])




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

import io
import os

class Set:

	"""
	Resource set. Implements :class:`Component`.
	
	:param Contextual source_path:
	   Source path.
	:param Contextual target_path:
	   Target path.
	:param Contextual props:
	   Resource properties.
	"""
	
	def __init__(self, source_path, target_path, props):
	
		self.__source_path = source_path
		self.__target_path = target_path
		self.__props = props
		
	def __repr__(self):
	
		return "{}.{}({}, {}, {})".format(
			self.__module__,
			type(self).__name__,
			repr(self.__source_path),
			repr(self.__target_path),
			repr(self.__props)
		)
		
	def __resolve(self, context, in_path, out_path):
	
		try:
			if os.path.basename(in_path)[0] != ".":
				if os.path.isfile(in_path):
					in_file = open(in_path, "r")
					out_file = open(out_path, "w")
					props = context.resolve(self.__props)
					resolver = Resolver(out_file, props)
					c = in_file.read(1)
					while len(c) > 0:
						resolver.update(c)
						c = in_file.read(1)
					in_file.close()
					out_file.close()
				elif os.path.isdir(in_path):
					if not os.path.exists(out_path):
						os.makedirs(out_path)
					for fname in os.listdir(in_path):
						self.__resolve(
							context,
							os.path.join(in_path, fname),
							os.path.join(out_path, fname)
						)
		except:
			print("Error resolving '{}'".format(in_path))
			raise
			
	def available(self, context):
	
		"""
		Return ``None``.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		return None
		
	def insert(self, context):
	
		"""
		Resolve resoures.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.__resolve(
			context.resolve(context, self.__source_path),
			context.resolve(context, self.__target_path)
		)
		
	def delete(self, context):
	
		"""
		Does nothing.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		pass
		
class Resolver:

	"""
	Resolve text resource variables.
	
	:param r_out:
	   Resolver output.
	:param dict r_vars:
	   Resolver variables. Variable with name ``context`` is reserved.
	"""
	
	def __init__(self, r_out, r_vars):
	
		self.__r_out = r_out
		self.__r_vars = r_vars
		self.__expr = io.StringIO()
		self.__update = self.__update_plain
		
	def __update_plain(self, c):
	
		if c == '#':
			self.__update = self.__update_sharp
		else:
			self.__r_out.write(c)
			
	def __update_sharp(self, c):
	
		if c == '{':
			self.__update = self.__update_expr
		else:
			if c == '#':
				self.__update = self.__update_sharpn
			else:
				self.__update = self.__update_plain
				self.__r_out.write('#')
			self.__r_out.write(c)
			
	def __update_sharpn(self, c):
	
		if c != '#':
			self.__update = self.__update_plain
		self.__r_out.write(c)
		
	def __update_expr(self, c):
	
		if c == '}':
			self.__update = self.__update_plain
			self.__resolve()
		else:
			if c == '\'':
				self.__update = self.__update_expr_quot
			self.__expr.write(c)
			
	def __update_expr_quot(self, c):
	
		if c == '\'':
			self.__update = self.__update_expr
		self.__expr.write(c)
		
	def __resolve(self):
	
		resolver = Resolver(self.__r_out, self.__r_vars)
		self.__expr.seek(0)
		for c in eval(self.__expr.read(), {}, self.__r_vars):
			resolver.update(c)
		self.__expr = io.StringIO()
		
	def update(self, c):
	
		"""
		Update output with next character.
		
		:param str c:
		   Next character.
		"""
		
		self.__update(c)




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import shutil
import subprocess

class Image:

	"""
	Packer image.
	
	:param Contextual base_dir:
	   Template base directory.
	:param Contextual template_path:
	   Template file path relative to ``base_dir``.
	"""
	
	def __init__(self, base_dir, template_path):
	
		self.__base_dir = base_dir
		self.__template_path = template_path
		
	def __repr__(self):
	
		return "{}.{}({}, {})".format(
			self.__module__,
			type(self).__name__,
			repr(self.__base_dir),
			repr(self.__template_path)
		)
		
	def available(self, context):
	
		"""
		Always return ``None``.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		return None
		
	def insert(self, context):
	
		"""
		Build the image.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		try:
			current_dir = os.getcwd()
			profile
			os.chdir(context.resolve(self.__base_dir))
			args = [
				"packer",
				"build",
				context.resolve(self.__template_path)
			]
			subprocess.call(args)
		finally:
			shutil.rmtree("output-*", ignore_errors=True)
			os.chdir(current_dir)
		
	def delete(self, context):
	
		"""
		Does nothing.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		pass




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess

class Project:

	"""
	Maven project. Implements :class:`Component`.
	
	:param Contextual base_dir:
	   Project base directory.
	:param bool dependency:
	   Project must be installed or only packaged.
	"""
	
	def __init__(self, base_dir, dependency=False):
	
		self.__base_dir = base_dir
		self.__dependency = dependency
		
	def __repr__(self):
	
		return "{}.{}({}, {})".format(
			self.__module__,
			type(self).__name__,
			repr(self.__base_dir)
		)
		
	def __execute(self, context, phase):
	
		subprocess.call([
			"mvn",
			"-f",
			os.path.join(os.path.join(context.resolve(self.__base_dir),
					"pom.xml")),
			phase
		])
		
	def available(self, context):
	
		"""
		Always return ``None``.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		return None
		
	def insert(self, context):
	
		"""
		Build the project.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		if self.__dependency:
			self.__execute(context, "install")
		else:
			self.__execute(context, "package")
		
	def delete(self, context):
	
		"""
		Does nothing.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		self.__execute(context, "clean")




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

from setuptools import setup, find_packages

import configparser
import os

package_ini = configparser.ConfigParser()
package_ini.read(os.path.join(
	os.path.dirname(os.path.abspath(__file__)),
	"package.ini"
))

setup(
	name="resort",
	version=package_ini["project"]["version"],
	
	author=package_ini["project"]["author"],
	author_email="miquel.ferran.gonzalez@gmail.com",
	
	packages=find_packages("packages"),
	package_dir={
		"": "packages"
	},
	install_requires={
		"colorama>=0.3.3"
	},
	test_suite="testsuite.resort",
	
	url="https://github.com/miquelo/resort",
	
	license="LICENSE.txt",
	description="Component manager for software projects.",
	long_description=(
		"Module for project management based on profiles whose components "
		"are included in dependency trees and execution plans. See "
		"documentation at http://miquelo.github.io/resort/."
	)
)




#
# This file is part of RESORT.
#
# RESORT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RESORT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RESORT.  If not, see <http://www.gnu.org/licenses/>.
#

import io
import os
import psycopg2

class Connection:

	"""
	PostgreSQL database connection.
	
	:param str host:
	   Database host.
	:param int port:
	   Database port.
	:param str dbname:
	   Database name.
	:param str username:
	   Database user name.
	:param str password:
	   Database password.
	"""
	
	def __init__(self, host, port, dbname, username, password):
	
		self.__host = host
		self.__port = port
		self.__dbname = dbname
		self.__username = username
		self.__password = password
		
	def __connect(self):
	
		return psycopg2.connect(
			host=self.__host,
			port=self.__port,
			database=self.__dbname,
			user=self.__username,
			password=self.__password,
			connect_timeout=20.
		)
		
	def execute(self, script):
	
		"""
		Execute script.
		
		:param str script:
		   Script to be executed.
		"""
		
		try:
			conn = self.__connect()
			conn.cursor().execute(script)
			conn.commit()
		finally:
			conn.close()
			
	def database_changes(self, script_path):
	
		"""
		Create a database changes component.
		
		:param Contextual script_path:
		   Path of script file.
		:rtype:
		   DatabaseChanges
		:return:
		   Database changes component.
		"""
		
		return DatabaseChanges(self, script_path)
		
class DatabaseChanges:

	"""
	Database changes. Implements :class:`Component`.
	
	:param Connection conn:
	   Target connection.
	:param Contextual script_path:
	   Script path.
	"""
	
	def __init__(self, conn, script_path):
	
		self.__conn = conn
		self.__script_path = script_path
		
	def __preprocess(self, src_path, out):
	
		exec_prefix = "exec '"
		base_dir = os.path.dirname(src_path)
		
		src_file = open(src_path, "r")
		for line in src_file.readlines():
			strip_line = line.strip()
			if strip_line.startswith(exec_prefix):
				exec_path = strip_line[len(exec_prefix):-1]
				self.__preprocess(os.path.join(base_dir, exec_path), out)
			elif len(strip_line) > 0:
				out.write("{}\n".format(strip_line))
		src_file.close()
		
	def available(self, context):
	
		"""
		Always returns ``None``.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		return None
		
	def insert(self, context):
	
		"""
		Applies database changes.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		script_path = context.resolve(self.__script_path)
		buf = io.StringIO()
		self.__preprocess(script_path, buf)
		buf.seek(0)
		self.__conn.execute(buf.read())
		
	def delete(self, context):
	
		"""
		Does nothing.
		
		:param resort.engine.execution.Context context:
		   Current execution context.
		"""
		
		pass

